import numpy as np
import matplotlib.pyplot as plt
import math
import Function

rng = np.random.default_rng(123)


def sigmoid(val):
    return 1 / (1 + np.exp(-val))


def sigmoid_derivative(val):
    return sigmoid(val) * (1 - sigmoid(val))


class NN:
    def __init__(self, input_size, hidden_size,
                 output_size, learning_rate, activation_function=sigmoid, derivative=sigmoid_derivative):
        self.input_size = input_size
        self.hidden_size = hidden_size  # This has to be a list of number of neurons in each hidden network
        self.hidden_layers = len(hidden_size)
        self.output_size = output_size
        self.activation_function = activation_function
        self.learning_rate = learning_rate
        self.derivative = derivative
        self.decay_rate = 0
        self.total_network = [self.input_size] + self.hidden_size + [self.output_size]
        self.weights = [[] for _ in range(self.hidden_layers + 1)]
        self.initialize_weights()
        self.outputs = [0] * self.output_size
        self.layer_outputs = []

    def initialize_weights(self):
        input_layer = True
        prev = self.input_size
        j = 0
        for n in self.total_network:
            if not input_layer:
                for i in range(n):
                    self.weights[j].append(rng.random(prev + 1))
                j = j + 1
            prev = n
            input_layer = False

    def forward_propagation(self, inputs):
        current = np.insert(inputs.copy().ravel(), 0, 1)
        for layer in self.weights:
            z = []
            for n in layer:
                n = np.array(n)
                z.append(self.activation_function(n.dot(current)))
            current = np.insert(z.copy(), 0, 1)
            self.layer_outputs.append(z)
        return z

    def int_to_bits(self, n):
        # Convert integer to binary string
        binary_str = bin(n)[2:]

        # Convert binary string to array of bits
        bit_array = [int(bit) for bit in binary_str]

        return bit_array

    def training(self, TEACHDATA):
        for epoch in TEACHDATA:
            for out, digit in enumerate(epoch):
                output_error = np.linalg.norm(
                    np.array(self.int_to_bits(out)) - np.array(self.forward_propagation(digit)))
                self.backpropagation(output_error)

    def backpropagation(self, output_error):
        for i in range(len(self.weights) - 1, -1, -1):
            z = []
            if i > 0:
                for n, w in enumerate(self.weights[i]):
                    w = np.array(w)
                    error_derivative = output_error * self.derivative(w.dot(np.insert(self.layer_outputs[i - 1], 0, 1)))
                    z.append(error_derivative)
            output_error = np.dot(np.array(self.weights[i]).T, np.array(z)[1:])
            self.weights[i] -= self.learning_rate * np.outer(z[1:], np.insert(self.layer_outputs[i], 0, 1))
