import numpy as np
import Training_data
import matplotlib.pyplot as plt
import Function

rng = np.random.default_rng(123)
TEACHDATA = 9999
TESTDATA = 999

T_NUMBER = 6  # Number to be detected 0-6

test_data = Training_data.make_testset(TESTDATA, 0.3)
teach_data = Training_data.make_testset(TEACHDATA, 0.2)


def sigmoid(val):
    return 1 / (1 + np.exp(-val))


def linear_act(val):
    # there is the option to make it linear in the valid range and make it constant otherwise
    return val


def threshold(val):
    return 1 if val > 0 else 0


class Neuron:

    def __init__(self, input_count, activation):
        self.input_count = input_count
        self.activation = activation
        self.weights = rng.random(input_count + 1)
        self.errors = []

    def plot_errors(self, titel):
        plt.plot(self.errors)
        plt.xlabel('Iteration')
        plt.ylabel('Error')
        plt.title(titel)
        plt.show()


class ThresholdPerceptron(Neuron):
    def __init__(self, input_count, activation=threshold):
        super().__init__(input_count, activation)

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            for j in rng.permutation(len(teach_data)):
                T = 1 if j == T_NUMBER else 0
                ix = np.insert(teach_data[j][i].ravel(), 0, 1)
                RI = self.activation(ix.dot(self.weights))
                if RI != T:
                    err = T - self.activation(ix.dot(self.weights))
                    delta = ETA * err * ix
                    self.weights = self.weights + delta
                    self.errors.append(abs(err))
            # if np.linalg.norm(old_weights - self.weights) == 0.00 and i % 57 == 0:
            #     return
        return

    def test(self, data):
        ix = np.insert(data.ravel(), 0, 1)
        return self.activation(ix.dot(self.weights))


class SGDPerceptron(Neuron):
    def __init__(self, input_count, activation=sigmoid):
        super().__init__(input_count, activation)

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            delta = [0 for _ in range(len(old_weights))]
            for j in rng.choice(rng.permutation(len(teach_data)), 5):
                T = (j == T_NUMBER)
                ix = np.insert(teach_data[j][i].ravel(), 0, 1)
                RI = self.activation(ix.dot(self.weights))
                err = T - RI
                delta = ETA * err * RI * (1 - RI) * ix
                self.errors.append(abs(err))
            self.weights += delta
            if np.linalg.norm(old_weights - self.weights) == 0.00:
                return
        return

    def test(self, data):
        ix = np.insert(data.ravel(), 0, 1)
        return self.activation(ix.dot(self.weights))


class LinearPerceptron(Neuron):
    def __init__(self, input_count, activation=linear_act):
        super().__init__(input_count, activation)

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            delta = [0 for _ in range(len(old_weights))]
            for j in rng.permutation(len(teach_data)):
                T = (j == T_NUMBER)
                ix = np.insert(teach_data[j][i].ravel(), 0, 1)
                err = T - self.activation(ix.dot(self.weights))
                delta += ETA * err * ix
                self.errors.append(abs(err))
            self.weights = self.weights + delta
            if np.linalg.norm(old_weights - self.weights) == 0.00:
                return
        return

    def test(self, data):
        ix = np.insert(data.ravel(), 0, 1)
        return self.activation(ix.dot(self.weights))


class Perceptron(Neuron):
    """This is using the one with the custom functions"""

    def __init__(self, input_count, activation):
        super().__init__(input_count, activation)
        self.derivative = activation.derivative('x').simplify()

    def train(self, ETA):
        for i in range(TEACHDATA):
            old_weights = np.copy(self.weights)
            delta = [0 for _ in range(len(old_weights))]
            for j in rng.choice(rng.permutation(len(teach_data)), 5):
                T = (j == T_NUMBER)
                ix = np.insert(teach_data[j][i].ravel(), 0, 1)
                err = T - self.activation.evaluate({'x': (ix.dot(self.weights))})
                delta = ETA * err * self.derivative.evaluate({'x': (ix.dot(self.weights))}) * ix
                self.errors.append(abs(err))
            self.weights += delta
            if np.linalg.norm(old_weights - self.weights) == 0.00:
                return
        return

    def test(self, data):
        ix = np.insert(data.ravel(), 0, 1)
        return self.activation.evaluate({'x': (ix.dot(self.weights))})
